searchData={"content_type":"text/markdown","items":[{"doc":"Provides macros for callback handling.","ref":"Exop.Callback.html","title":"Exop.Callback","type":"behaviour"},{"doc":"defmodule CallbackModule do\n      use Exop.Callback\n\n      def process(operation, params, result, opts) do\n        # your callback handling code here\n      end\n    end\n\n    defmodule TestOperation do\n      use Exop.Operation\n\n      callback CallbackModule, return: true\n\n      parameter :a, type: :integer\n      parameter :b, type: :integer\n\n      def process(%{a: a, b: b}), do: a + b\n    end\n\nBy using `Exop.Callback` you need to implement `process/4` function which takes following params:\n  * operation module (`TestOperation` in the example above)\n  * params that were passed into the operation (`%{a: 1, b: 2}`)\n  * an successful result which was returned by the operation.\n  * opts is an open keyword list for metadata passing.\n\nCallback operation has no effect on the result. but trigger a callback after success coming handy to send\nPubSub prodcast or any side effect event without interrepting the main workflow.","ref":"Exop.Callback.html#module-example","title":"Example - Exop.Callback","type":"behaviour"},{"doc":"Callback handling function.\nReceives:\n  - operation module\n  - a map of parameters with which the operation was invoked\n  - an result tuple returned by the operation\n  - a keyword list defined as opts in within macro call.","ref":"Exop.Callback.html#c:process/4","title":"Exop.Callback.process/4","type":"callback"},{"doc":"","ref":"Exop.Callback.html#t:t/0","title":"Exop.Callback.t/0","type":"type"},{"doc":"Provides macros to organize a number of Exop.Operation modules into an invocation chain.","ref":"Exop.Chain.html","title":"Exop.Chain","type":"module"},{"doc":"defmodule CreateUser do\n      use Exop.Chain\n\n      alias Operations.{User, Backoffice, Notifications}\n\n      operation User.Create\n      operation Backoffice.SaveStats\n      operation Notifications.SendEmail\n    end\n\n    # CreateUser.run(name: \"User Name\", age: 37, gender: \"m\")\n\n`Exop.Chain` defines `run/1` function that takes `keyword()` or `map()` of params.\nThose params will be passed into the first operation in the chain.\nBear in mind that each of chained operations (except the first one) awaits a returned result of\na previous operation as incoming params.\n\nSo in the example above `CreateUser.run(name: \"User Name\", age: 37, gender: \"m\")` will invoke\nthe chain by passing `[name: \"User Name\", age: 37, gender: \"m\"]` params to the first `User.Create`\noperation.\nThe result of `User.Create` operation will be passed to `Backoffice.SaveStats`\noperation as its params and so on.\n\nOnce any of operations in the chain returns non-ok-tuple result (error result, interruption, auth error etc.)\nthe chain execution interrupts and error result returned (as the chain (`CreateUser`) result).","ref":"Exop.Chain.html#module-example","title":"Example - Exop.Chain","type":"module"},{"doc":"Defines one of a chain's operation.","ref":"Exop.Chain.html#operation/2","title":"Exop.Chain.operation/2","type":"macro"},{"doc":"Defines one of a chain's operation.","ref":"Exop.Chain.html#step/2","title":"Exop.Chain.step/2","type":"macro"},{"doc":"Provides macros for fallback handling.","ref":"Exop.Fallback.html","title":"Exop.Fallback","type":"behaviour"},{"doc":"defmodule FallbackModule do\n      use Exop.Fallback\n\n      def process(operation, params, error) do\n        # your error handling code here\n        :some_fallback_result\n      end\n    end\n\n    defmodule TestOperation do\n      use Exop.Operation\n\n      fallback FallbackModule, return: true\n\n      parameter :a, type: :integer\n      parameter :b, type: :integer\n\n      def process(%{a: a, b: b}), do: a + b\n    end\n\n    # TestOperation.run(a: 1, b: \"2\") => :some_fallback_result\n\nBy using `Exop.Fallback` you need to implement `process/3` function which takes following params:\n  * failed operation module (`TestOperation` in the example above)\n  * params that were passed into the operation (`%{a: 1, b: \"2\"}`)\n  * an error result which was returned by the operation (`{:error, {:validation, %{a: [\"has wrong type\"]}}}`)\n\nDuring a fallback definition you can add `return: true` option so in the example case\n`TestOperation.run/1` will return the result of the fallback (`FallbackModule.process/3`\n function's result - `:some_fallback_result`).\nIf you want `TestOperation.run/1` to return original result (`{:error, {:validation, %{a: [\"has wrong type\"]}}}`)\nspecify `return: false` option or just omit it in a fallback definition.","ref":"Exop.Fallback.html#module-example","title":"Example - Exop.Fallback","type":"behaviour"},{"doc":"Fallback handling function.\nReceives:\n  - failed operation module\n  - a map of parameters with which the operation was invoked\n  - an error tuple returned by the operation","ref":"Exop.Fallback.html#c:process/3","title":"Exop.Fallback.process/3","type":"callback"},{"doc":"","ref":"Exop.Fallback.html#t:t/0","title":"Exop.Fallback.t/0","type":"type"},{"doc":"Provides macros for an operation's contract definition and process/1 function.","ref":"Exop.Operation.html","title":"Exop.Operation","type":"behaviour"},{"doc":"defmodule SomeOperation do\n      use Exop.Operation\n\n      parameter :param1, type: :integer, required: true\n      parameter :param2, type: :string, length: %{max: 3}, format: ~r/foo/\n\n      def process(params) do\n        \"This is the operation's result with one of the params = \"   params[:param1]\n      end\n    end","ref":"Exop.Operation.html#module-example","title":"Example - Exop.Operation","type":"behaviour"},{"doc":"Authorizes an action with predefined policy (see `Policy check` macro docs).\nIf authorization fails, any code after (below) auth check will be postponed (an error `{:error, {:auth, _reason}}` will be returned immediately)","ref":"Exop.Operation.html#authorize/1","title":"Exop.Operation.authorize/1","type":"macro"},{"doc":"Defines a callback module that will be used for a successful operation as side effect.\n    defmodule MultiplyByTenOperation do\n      use Exop.Operation\n\n      callback LiveProdcast, topic: \"math\", type: \"mutliplication\"\n\n      parameter :a, type: :integer, required: true\n\n      def process(%{a: a}), do: a * 10\n    end\n\nA callback module itself might be:\n    defmodule LiveProdcast do\n      use Exop.Callback\n\n      def process(operation_module, params_passed_to_the_operation, successful_result, opts) do\n        Phoenix.PubSub.broadcast(MyApp.PubSub, opts[:topic], %{type: opts[:type], payload: successful_result})\n      end\n    end\n\n`opts` is an open keyword list to pass needed options and metadata to the successful call.","ref":"Exop.Operation.html#callback/2","title":"Exop.Operation.callback/2","type":"macro"},{"doc":"Returns policy that was defined in an operation.","ref":"Exop.Operation.html#current_policy/0","title":"Exop.Operation.current_policy/0","type":"macro"},{"doc":"Defines a fallback module that will be used for an operation's non-ok-tuple (fail) result handling.\n    defmodule MultiplyByTenOperation do\n      use Exop.Operation\n\n      fallback LoggerFallback\n\n      parameter :a, type: :integer, required: true\n\n      def process(%{a: a}), do: a * 10\n    end\n\nA fallback module itself might be:\n    defmodule LoggerFallback do\n      use Exop.Fallback\n      require Logger\n\n      def process(operation_module, params_passed_to_the_operation, operation_error_result) do\n        Logger.error(\"Oops\")\n      end\n    end\n\nIf `return: true` option is provided then failed operation's `run/1` will return the\nfallback's `process/3` result.","ref":"Exop.Operation.html#fallback/2","title":"Exop.Operation.fallback/2","type":"macro"},{"doc":"Defines a parameter with `name` and `opts` in an operation contract.\nOptions could include the parameter value checks and transformations (like coercion).\n\nA parameter name could be either an atom or a string. You could even mix atom-named and\nstring-named parameters in an operation's contract.","ref":"Exop.Operation.html#parameter/2","title":"Exop.Operation.parameter/2","type":"macro"},{"doc":"parameter :some_param, type: :map, required: true\n    parameter \"my parameter\", type: :map, required: true","ref":"Exop.Operation.html#parameter/2-example","title":"Example - Exop.Operation.parameter/2","type":"macro"},{"doc":"##","ref":"Exop.Operation.html#parameter/2-available-checks-are","title":"Available checks are: - Exop.Operation.parameter/2","type":"macro"},{"doc":"Checks whether a parameter's value is of declared type.\n    parameter :some_param, type: :map\n\n##","ref":"Exop.Operation.html#parameter/2-type","title":"type - Exop.Operation.parameter/2","type":"macro"},{"doc":"Checks the presence of a parameter in passed params collection.\n    parameter :some_param, required: true\n\n##","ref":"Exop.Operation.html#parameter/2-required","title":"required - Exop.Operation.parameter/2","type":"macro"},{"doc":"Checks if the parameter is missed and assigns default value to it if so.\n    parameter :some_param, default: \"default value\"\n\n##","ref":"Exop.Operation.html#parameter/2-default","title":"default - Exop.Operation.parameter/2","type":"macro"},{"doc":"Checks whether a parameter's value is a number and passes constraints (if constraints were defined).\n    parameter :some_param, numericality: %{equal_to: 10, greater_than: 0,\n                                           greater_than_or_equal_to: 10,\n                                           less_than: 20,\n                                           less_than_or_equal_to: 10}\n\n##","ref":"Exop.Operation.html#parameter/2-numericality","title":"numericality - Exop.Operation.parameter/2","type":"macro"},{"doc":"Checks whether a parameter's value exactly equals given value (with type equality).\n    parameter :some_param, equals: 100.5\n\n##","ref":"Exop.Operation.html#parameter/2-equals","title":"equals - Exop.Operation.parameter/2","type":"macro"},{"doc":"Checks whether a parameter's value is within a given list.\n    parameter :some_param, in: ~w(a b c)\n\n##","ref":"Exop.Operation.html#parameter/2-in","title":"in - Exop.Operation.parameter/2","type":"macro"},{"doc":"Checks whether a parameter's value is not within a given list.\n    parameter :some_param, not_in: ~w(a b c)\n\n##","ref":"Exop.Operation.html#parameter/2-not_in","title":"not_in - Exop.Operation.parameter/2","type":"macro"},{"doc":"Checks wether parameter's value matches given regex.\n    parameter :some_param, format: ~r/foo/\n\n##","ref":"Exop.Operation.html#parameter/2-format","title":"format - Exop.Operation.parameter/2","type":"macro"},{"doc":"Checks the length of a parameter's value.\n    parameter :some_param, length: %{min: 5, max: 10, is: 7, in: 5..8}\n\n##","ref":"Exop.Operation.html#parameter/2-length","title":"length - Exop.Operation.parameter/2","type":"macro"},{"doc":"Checks the inner of either Map or Keyword parameter.\n    parameter :some_param, type: :map, inner: %{\n      a: [type: :integer, required: true],\n      b: [type: :string, length: %{min: 1, max: 6}]\n    }\n\n##","ref":"Exop.Operation.html#parameter/2-inner","title":"inner - Exop.Operation.parameter/2","type":"macro"},{"doc":"Checks whether the given parameter is expected structure.\n    parameter :some_param, struct: %SomeStruct{}\n\n##","ref":"Exop.Operation.html#parameter/2-struct","title":"struct - Exop.Operation.parameter/2","type":"macro"},{"doc":"Checks whether each of list items conforms defined checks. An item's checks could be any that Exop offers:\n    parameter :list_param, list_item: %{type: :string, length: %{min: 7}}\n\n##","ref":"Exop.Operation.html#parameter/2-list_item","title":"list_item - Exop.Operation.parameter/2","type":"macro"},{"doc":"Checks whether an item is valid over custom validation function.\n    parameter :some_param, func: &__MODULE__.your_validation/2\n\n    def your_validation({param_name, param_value}, all_received_params_map) do\n      # your validation logic based on given arguments is here\n    end\n\n##","ref":"Exop.Operation.html#parameter/2-func","title":"func - Exop.Operation.parameter/2","type":"macro"},{"doc":"It is not a parameter check itself, because it doesn't return any validation errors.\nIt is a parameter attribute which allow you to have other checks for a parameter whilst have\na possibility to pass `nil` as the parameter's value.\nIf `nil` is passed all the parameter's checks are ignored during validation.\n\n##","ref":"Exop.Operation.html#parameter/2-allow_nil","title":"allow_nil - Exop.Operation.parameter/2","type":"macro"},{"doc":"This option allows you to pass a parameter to `run/1` and `run!/1` functions with one name and\nwork with this parameter within an operation under another name.\n\n    parameter :a, type: :integer, from: \"a\"\n\n##","ref":"Exop.Operation.html#parameter/2-from","title":"from - Exop.Operation.parameter/2","type":"macro"},{"doc":"Checks whether a parameter's value (list) is a subset of a defined check-list.\nTo pass this check, all items within given into an operation parameter should be included\ninto check-list, otherwise the check is failed.\n\n    parameter :some_param, subset_of: [1, 2, :a, \"b\", C]","ref":"Exop.Operation.html#parameter/2-subset_of","title":"subset_of - Exop.Operation.parameter/2","type":"macro"},{"doc":"In some cases you might want to make an 'early return' from `process/1` function.\nFor this purpose you can call `interrupt/1` function within `process/1` and pass an interruption reason to it.\nAn operation will be interrupted and return `{:interrupt, your_reason}`\n\n    def process(_params) do\n      interrupt(%{fail: \"oops\"})\n      :ok # will not return it\n    end","ref":"Exop.Operation.html#parameter/2-interrupt","title":"Interrupt - Exop.Operation.parameter/2","type":"macro"},{"doc":"It is possible to coerce a parameter before the contract validation, all validation checks\nwill be invoked on coerced parameter value.\nSince coercion changes a parameter before any validation has been invoked,\ndefault values are resolved (with `:default` option) before the coercion.\nThe flow looks like: `Resolve param default value -> Coerce -> Validate coerced`\n\n    parameter :a, type: :string, coerce_with: &__MODULE__.to_string/2\n\n    def to_string({:a, value}, %{} = _received_params) when is_integer(value) do\n      Integer.to_string(value)\n    end\n    def to_string({:a, value}, %{} = _received_params) when is_binary(value) do\n      value\n    end\n\n_For more information and examples check out general Exop docs._","ref":"Exop.Operation.html#parameter/2-coercion","title":"Coercion - Exop.Operation.parameter/2","type":"macro"},{"doc":"Defines a policy that will be used for authorizing the possibility of a user\nto invoke an operation.\n    defmodule ReadOperation do\n      use Exop.Operation\n\n      policy MonthlyReportPolicy, :can_read?\n\n      parameter :user, required: true, struct: User\n\n      def process(params) do\n        authorize(params.user)\n\n        # make some reading...\n      end\n    end\n\nA policy itself might be:\n    defmodule MonthlyReportPolicy do\n      # not only Keyword or Map as an argument since 1.1.1\n      def can_read?(%User{role: \"manager\"}), do: true\n      def can_read?(_opts), do: false\n\n      def can_write?(%User{role: \"manager\"}), do: true\n      def can_write?(_opts), do: false\n    end","ref":"Exop.Operation.html#policy/2","title":"Exop.Operation.policy/2","type":"macro"},{"doc":"Operation's entry point. Takes defined contract as the single parameter.\nContract itself is a list of maps: `[%{name: atom(), opts: keyword()}]`","ref":"Exop.Operation.html#c:process/1","title":"Exop.Operation.process/1","type":"callback"},{"doc":"String.Chars protocol implementations for some types","ref":"Exop.StringCharsImplementations.html","title":"Exop.StringCharsImplementations","type":"module"},{"doc":"","ref":"Exop.TypeValidation.html","title":"Exop.TypeValidation","type":"module"},{"doc":"","ref":"Exop.TypeValidation.html#check_value/2","title":"Exop.TypeValidation.check_value/2","type":"function"},{"doc":"","ref":"Exop.TypeValidation.html#known_types/0","title":"Exop.TypeValidation.known_types/0","type":"function"},{"doc":"","ref":"Exop.TypeValidation.html#type_supported?/2","title":"Exop.TypeValidation.type_supported?/2","type":"function"},{"doc":"A bunch of common functions.","ref":"Exop.Utils.html","title":"Exop.Utils","type":"module"},{"doc":"","ref":"Exop.Utils.html#defined_params/2","title":"Exop.Utils.defined_params/2","type":"function"},{"doc":"","ref":"Exop.Utils.html#fetch_inner_checks/1","title":"Exop.Utils.fetch_inner_checks/1","type":"function"},{"doc":"","ref":"Exop.Utils.html#put_param_value/3","title":"Exop.Utils.put_param_value/3","type":"function"},{"doc":"","ref":"Exop.Utils.html#resolve_coercions/3","title":"Exop.Utils.resolve_coercions/3","type":"function"},{"doc":"","ref":"Exop.Utils.html#resolve_defaults/3","title":"Exop.Utils.resolve_defaults/3","type":"function"},{"doc":"","ref":"Exop.Utils.html#resolve_from/3","title":"Exop.Utils.resolve_from/3","type":"function"},{"doc":"Tries to make a map from a struct and keyword list","ref":"Exop.Utils.html#try_map/1","title":"Exop.Utils.try_map/1","type":"function"},{"doc":"Provides high-level functions for a contract validation.\n  The main function is valid?/2\n  Mostly invokes Exop.ValidationChecks module functions.","ref":"Exop.Validation.html","title":"Exop.Validation","type":"module"},{"doc":"Checks inner item of the contract param (which is a Map itself) with their own checks.","ref":"Exop.Validation.html#check_inner/3","title":"Exop.Validation.check_inner/3","type":"function"},{"doc":"iex> Exop.Validation.check_inner(%{a: 1}, :a, [b: [type: :atom], c: [type: :string]])\n    [%{a: \"has wrong type\"}]\n\n    iex> Exop.Validation.check_inner(%{a: []}, :a, [b: [type: :atom], c: [type: :string]])\n    [[%{\"a[:b]\" => \"is required\"}, true], [%{\"a[:c]\" => \"is required\"}, true]]\n\n    iex> Exop.Validation.check_inner(%{a: %{b: :atom, c: \"string\"}}, :a, [b: [type: :atom], c: [type: :string]])\n    [[true, true], [true, true]]","ref":"Exop.Validation.html#check_inner/3-examples","title":"Examples - Exop.Validation.check_inner/3","type":"function"},{"doc":"Checks items of the contract's list param with specified checks.","ref":"Exop.Validation.html#check_list_item/3","title":"Exop.Validation.check_list_item/3","type":"function"},{"doc":"iex> Exop.Validation.check_list_item(%{a: 1}, :a, [type: :integer])\n    [%{a: \"is not a list\"}]\n\n    iex> Exop.Validation.check_list_item(%{a: []}, :a, [type: :integer])\n    []\n\n    iex> Exop.Validation.check_list_item(%{a: [1, :atom]}, :a, [type: :integer])\n    [[true, true], [true, %{\"a[1]\" => \"has wrong type; expected type: integer, got: :atom\"}]]\n\n    iex> Exop.Validation.check_list_item(%{a: [1, 2]}, :a, [type: :integer])\n    [[true, true], [true, true]]","ref":"Exop.Validation.html#check_list_item/3-examples","title":"Examples - Exop.Validation.check_list_item/3","type":"function"},{"doc":"","ref":"Exop.Validation.html#errors_message/1","title":"Exop.Validation.errors_message/1","type":"function"},{"doc":"Validate received params over a contract.","ref":"Exop.Validation.html#valid?/2","title":"Exop.Validation.valid?/2","type":"function"},{"doc":"iex> Exop.Validation.valid?([%{name: :param, opts: [required: true]}], [param: \"hello\"])\n    :ok","ref":"Exop.Validation.html#valid?/2-examples","title":"Examples - Exop.Validation.valid?/2","type":"function"},{"doc":"Validate received params over a contract. Accumulate validation results into a list.","ref":"Exop.Validation.html#validate/3","title":"Exop.Validation.validate/3","type":"function"},{"doc":"iex> Exop.Validation.validate([%{name: :param, opts: [required: true, type: :string]}], [param: \"hello\"], [])\n    [true, true]","ref":"Exop.Validation.html#validate/3-examples","title":"Examples - Exop.Validation.validate/3","type":"function"},{"doc":"","ref":"Exop.Validation.html#t:validation_error/0","title":"Exop.Validation.validation_error/0","type":"type"},{"doc":"Provides low-level validation functions:\n\n  * check_type/3\n  * check_required/3\n  * check_numericality/3\n  * check_in/3\n  * check_not_in/3\n  * check_format/3\n  * check_length/3\n  * check_struct/3\n  * check_func/3\n  * check_equals/3\n  * check_exactly/3\n  * check_allow_nil/3","ref":"Exop.ValidationChecks.html","title":"Exop.ValidationChecks","type":"module"},{"doc":"","ref":"Exop.ValidationChecks.html#check_allow_nil/3","title":"Exop.ValidationChecks.check_allow_nil/3","type":"function"},{"doc":"Checks whether a parameter's value exactly equals given value (with type equality).","ref":"Exop.ValidationChecks.html#check_equals/3","title":"Exop.ValidationChecks.check_equals/3","type":"function"},{"doc":"iex> Exop.ValidationChecks.check_equals(%{a: 1}, :a, 1)\n    true","ref":"Exop.ValidationChecks.html#check_equals/3-examples","title":"Examples - Exop.ValidationChecks.check_equals/3","type":"function"},{"doc":"The alias for `check_equals/3`.\nChecks whether a parameter's value exactly equals given value (with type equality).","ref":"Exop.ValidationChecks.html#check_exactly/3","title":"Exop.ValidationChecks.check_exactly/3","type":"function"},{"doc":"iex> Exop.ValidationChecks.check_exactly(%{a: 1}, :a, 1)\n    true","ref":"Exop.ValidationChecks.html#check_exactly/3-examples","title":"Examples - Exop.ValidationChecks.check_exactly/3","type":"function"},{"doc":"Checks whether an item_name conforms the given format.","ref":"Exop.ValidationChecks.html#check_format/3","title":"Exop.ValidationChecks.check_format/3","type":"function"},{"doc":"iex> Exop.ValidationChecks.check_format(%{a: \"bar\"}, :a, ~r/bar/)\n    true","ref":"Exop.ValidationChecks.html#check_format/3-examples","title":"Examples - Exop.ValidationChecks.check_format/3","type":"function"},{"doc":"Checks whether an item is valid over custom validation function.","ref":"Exop.ValidationChecks.html#check_func/3","title":"Exop.ValidationChecks.check_func/3","type":"function"},{"doc":"iex> Exop.ValidationChecks.check_func(%{a: 1}, :a, fn({:a, value}, _all_param)-> value > 0 end)\n    true\n    iex> Exop.ValidationChecks.check_func(%{a: 1}, :a, fn({:a, _value}, _all_param)-> :ok end)\n    true\n    iex> Exop.ValidationChecks.check_func(%{a: 1}, :a, fn({:a, value}, _all_param)-> is_nil(value) end)\n    %{a: \"not valid\"}\n    iex> Exop.ValidationChecks.check_func(%{a: 1}, :a, fn({:a, _value}, _all_param)-> :error end)\n    %{a: \"not valid\"}\n    iex> Exop.ValidationChecks.check_func(%{a: -1}, :a, fn({:a, _value}, _all_param)-> {:error, :my_error} end)\n    %{a: :my_error}","ref":"Exop.ValidationChecks.html#check_func/3-examples","title":"Examples - Exop.ValidationChecks.check_func/3","type":"function"},{"doc":"Checks whether an item_name is a memeber of a list.","ref":"Exop.ValidationChecks.html#check_in/3","title":"Exop.ValidationChecks.check_in/3","type":"function"},{"doc":"iex> Exop.ValidationChecks.check_in(%{a: 1}, :a, [1, 2, 3])\n    true","ref":"Exop.ValidationChecks.html#check_in/3-examples","title":"Examples - Exop.ValidationChecks.check_in/3","type":"function"},{"doc":"Checks whether a check_item has been provided.\nReturns a boolean.","ref":"Exop.ValidationChecks.html#check_item_present?/2","title":"Exop.ValidationChecks.check_item_present?/2","type":"function"},{"doc":"iex> Exop.ValidationChecks.check_item_present?(%{a: 1, b: 2}, :a)\n    true\n\n    iex> Exop.ValidationChecks.check_item_present?([a: 1, b: 2], :b)\n    true\n\n    iex> Exop.ValidationChecks.check_item_present?(%{a: 1, b: 2}, :c)\n    false\n\n    iex> Exop.ValidationChecks.check_item_present?(%{a: 1, b: nil}, :b)\n    true","ref":"Exop.ValidationChecks.html#check_item_present?/2-examples","title":"Examples - Exop.ValidationChecks.check_item_present?/2","type":"function"},{"doc":"Checks an item_name over length constraints.","ref":"Exop.ValidationChecks.html#check_length/3","title":"Exop.ValidationChecks.check_length/3","type":"function"},{"doc":"iex> Exop.ValidationChecks.check_length(%{a: \"123\"}, :a, %{min: 0})\n    [true]\n\n    iex> Exop.ValidationChecks.check_length(%{a: ~w(1 2 3)}, :a, %{in: 2..4})\n    [true]\n\n    iex> Exop.ValidationChecks.check_length(%{a: ~w(1 2 3)}, :a, %{is: 3, max: 4})\n    [true, true]","ref":"Exop.ValidationChecks.html#check_length/3-examples","title":"Examples - Exop.ValidationChecks.check_length/3","type":"function"},{"doc":"Checks whether an item_name is not a memeber of a list.","ref":"Exop.ValidationChecks.html#check_not_in/3","title":"Exop.ValidationChecks.check_not_in/3","type":"function"},{"doc":"iex> Exop.ValidationChecks.check_not_in(%{a: 4}, :a, [1, 2, 3])\n    true","ref":"Exop.ValidationChecks.html#check_not_in/3-examples","title":"Examples - Exop.ValidationChecks.check_not_in/3","type":"function"},{"doc":"Checks an item_name over numericality constraints.","ref":"Exop.ValidationChecks.html#check_numericality/3","title":"Exop.ValidationChecks.check_numericality/3","type":"function"},{"doc":"iex> Exop.ValidationChecks.check_numericality(%{a: 3}, :a, %{ equal_to: 3 })\n    true\n\n    iex> Exop.ValidationChecks.check_numericality(%{a: 5}, :a, %{ greater_than_or_equal_to: 3 })\n    true\n\n    iex> Exop.ValidationChecks.check_numericality(%{a: 3}, :a, %{ less_than_or_equal_to: 3 })\n    true","ref":"Exop.ValidationChecks.html#check_numericality/3-examples","title":"Examples - Exop.ValidationChecks.check_numericality/3","type":"function"},{"doc":"The alias for `check_format/3`.\nChecks whether an item_name conforms the given format.","ref":"Exop.ValidationChecks.html#check_regex/3","title":"Exop.ValidationChecks.check_regex/3","type":"function"},{"doc":"iex> Exop.ValidationChecks.check_regex(%{a: \"bar\"}, :a, ~r/bar/)\n    true","ref":"Exop.ValidationChecks.html#check_regex/3-examples","title":"Examples - Exop.ValidationChecks.check_regex/3","type":"function"},{"doc":"Checks if an item_name presents in params if its required (true).","ref":"Exop.ValidationChecks.html#check_required/3","title":"Exop.ValidationChecks.check_required/3","type":"function"},{"doc":"iex> Exop.ValidationChecks.check_required(%{}, :some_item, false)\n    true\n\n    iex> Exop.ValidationChecks.check_required([a: 1, b: 2], :a, true)\n    true\n\n    iex> Exop.ValidationChecks.check_required(%{a: 1, b: 2}, :b, true)\n    true","ref":"Exop.ValidationChecks.html#check_required/3-examples","title":"Examples - Exop.ValidationChecks.check_required/3","type":"function"},{"doc":"Checks whether an item is expected structure.","ref":"Exop.ValidationChecks.html#check_struct/3","title":"Exop.ValidationChecks.check_struct/3","type":"function"},{"doc":"defmodule SomeStruct1, do: defstruct [:a, :b]\n    defmodule SomeStruct2, do: defstruct [:b, :c]\n\n    Exop.ValidationChecks.check_struct(%{a: %SomeStruct1{}}, :a, %SomeStruct1{})\n    # true\n    Exop.ValidationChecks.check_struct(%{a: %SomeStruct1{}}, :a, %SomeStruct2{})\n    # false","ref":"Exop.ValidationChecks.html#check_struct/3-examples","title":"Examples - Exop.ValidationChecks.check_struct/3","type":"function"},{"doc":"","ref":"Exop.ValidationChecks.html#check_subset_of/3","title":"Exop.ValidationChecks.check_subset_of/3","type":"function"},{"doc":"Checks the type of an item_name.","ref":"Exop.ValidationChecks.html#check_type/3","title":"Exop.ValidationChecks.check_type/3","type":"function"},{"doc":"iex> Exop.ValidationChecks.check_type(%{a: 1}, :a, :integer)\n    true\n\n    iex> Exop.ValidationChecks.check_type(%{a: \"1\"}, :a, :string)\n    true\n\n    iex> Exop.ValidationChecks.check_type(%{a: nil}, :a, :string)\n    %{:a => \"has wrong type; expected type: string, got: nil\"}","ref":"Exop.ValidationChecks.html#check_type/3-examples","title":"Examples - Exop.ValidationChecks.check_type/3","type":"function"},{"doc":"Returns an check_item's value from either a Keyword or a Map by an atom-key.","ref":"Exop.ValidationChecks.html#get_check_item/2","title":"Exop.ValidationChecks.get_check_item/2","type":"function"},{"doc":"iex> Exop.ValidationChecks.get_check_item(%{a: 1, b: 2}, :a)\n    1\n\n    iex> Exop.ValidationChecks.get_check_item([a: 1, b: 2], :b)\n    2\n\n    iex> Exop.ValidationChecks.get_check_item(%{a: 1, b: 2}, :c)\n    nil","ref":"Exop.ValidationChecks.html#get_check_item/2-examples","title":"Examples - Exop.ValidationChecks.get_check_item/2","type":"function"},{"doc":"","ref":"Exop.ValidationChecks.html#t:check_error/0","title":"Exop.ValidationChecks.check_error/0","type":"type"},{"doc":"","ref":"Exop.Operation.ErrorResult.html","title":"Exop.Operation.ErrorResult","type":"exception"},{"doc":"An operation's contract validation failure error.","ref":"Exop.Validation.ValidationError.html","title":"Exop.Validation.ValidationError","type":"exception"},{"doc":"[![Hex.pm](https://img.shields.io/hexpm/v/exop.svg)](https://hex.pm/packages/exop) [![API Docs](https://img.shields.io/badge/api-docs-yellow.svg?style=flat)](http://hexdocs.pm/exop/) [![Build Status](https://travis-ci.org/madeinussr/exop.svg?branch=master)](https://travis-ci.org/madeinussr/exop)\n\n# Exop\n\nA library that helps you to organize your Elixir code in more domain-driven way.\nExop provides macros which helps you to encapsulate business logic and offers you additionally:\nincoming params validation (with predefined contract), params coercion, policy check, fallback behavior, operations chaining and more.\n\n---","ref":"readme.html","title":"README","type":"extras"},{"doc":"#","ref":"readme.html#exop-family","title":"Exop family: - README","type":"extras"},{"doc":"Interested in property-based testing? Check out new Exop family member - [ExopData](https://github.com/madeinussr/exop_data). If you use Exop to organize your code with ExopData you can get property generators in the most easiest way.\n\n#","ref":"readme.html#exopdata","title":"ExopData - README","type":"extras"},{"doc":"The new [ExopPlug](https://github.com/madeinussr/exop_plug) library provides a convenient way to validate incoming parameters of your Phoenix application's controllers by offering you small but useful DSL.\n\n---","ref":"readme.html#exopplug","title":"ExopPlug - README","type":"extras"},{"doc":"Here is the [CHANGELOG](https://github.com/madeinussr/exop/blob/master/CHANGELOG.md) that was started from ver. 0.4.1 ¯\\\\\\_(ツ)\\_/¯\n\n-   [Installation](#installation)\n-   [Operation definition](#operation-definition)\n    -   [Parameter checks](#parameter-checks)\n        -   [type](#type)\n        -   [required](#required)\n        -   [default](#default)\n        -   [numericality](#numericality)\n        -   [equals](#equals)\n        -   [in](#in)\n        -   [not_in](#not_in)\n        -   [format](#format)\n        -   [length](#length)\n        -   [inner](#inner)\n        -   [struct](#struct)\n        -   [list_item](#list_item)\n        -   [func](#func)\n        -   [allow_nil](#allow_nil)\n        -   [from](#from)\n        -   [subset_of](#subset_of)\n    -   [Defined params](#defined-params)\n    -   [Interrupt](#interrupt)\n    -   [Coercion](#coercion)\n    -   [Policy check](#policy-check)\n    -   [Fallback module](#fallback-module)\n    -   [Callback module](#callback-module)\n-   [Operation invocation](#operation-invocation)\n-   [Operation results](#operation-results)\n-   [Operations chain](#operations-chain)\n    -   [Additional parameters](#additional-parameters)\n    -   [Descriptive errors](#descriptive-errors)\n    -   [Conditional operations](#conditional-operations)\n    -   [Incoming parameters coercion](#incoming-parameters-coercion)","ref":"readme.html#table-of-contents","title":"Table of Contents - README","type":"extras"},{"doc":"```elixir\ndef deps do\n  [{:exop, \"~> 1.4\"}]\nend\n```","ref":"readme.html#installation","title":"Installation - README","type":"extras"},{"doc":"```elixir\ndefmodule IntegersDivision do\n  use Exop.Operation\n\n  parameter :a, type: :integer, default: 1\n  parameter :b, type: :integer, required: false,\n                numericality: %{greater_than: 0}\n\n  def process(params) do\n    result = params[:a] / params[:b]\n    IO.inspect \"The division result is: #{result}\"\n  end\nend\n```\n\n`Exop.Operation` provides `parameter` macro, which is responsible for the contract definition.\nIts spec is `@spec parameter(atom | String.t, Keyword.t) :: none`, we define parameter name as the first argument and parameter options as the second `Keyword` argument.\n\n_A parameter name could be either an atom or a string. You could even mix atom-named and string-named parameters in an operation's contract._\n\nParameter options determine a contract of a parameter, a set of parameters contracts is an operation contract.\n\nBusiness logic of an operation is defined in `process/1` function, which is required by the Exop.Operation module\nbehaviour.\n\nAfter the contract and business logic were defined, you can invoke the operation simply by calling `run/1` function:\n\n```elixir\niex> IntegersDivision.run(a: 50, b: 5)\n{:ok, \"The division result is: 10\"}\n```\n\nReturn type will be either `{:ok, any()}` (where the second item in the tuple is `process/1` function's result) or\n`{:error, {:validation, map()}}` (where the `map()` is validation errors map).\n\n_for more information see [Operation results](#operation-results) section_\n\n#","ref":"readme.html#operation-definition","title":"Operation definition - README","type":"extras"},{"doc":"A parameter options could have various checks. Here the list of available checks:\n\n-   `type`\n-   `required`\n-   `default`\n-   `numericality`\n-   `equals` (`exactly`)\n-   `in`\n-   `not_in`\n-   `format` (`regex`)\n-   `length`\n-   `inner`\n-   `struct`\n-   `list_item`\n-   `func`\n-   `allow_nil`\n-   `from`\n-   `subset_of`\n\n#### `type`\n\nChecks whether a parameter's value is of declared type.\n\n```elixir\nparameter :some_param, type: :map\n```\n\nExop handle almost all Elixir types and some additional:\n\n-   :boolean\n-   :integer\n-   :float\n-   :string\n-   :tuple\n-   :map\n-   :keyword\n-   :list\n-   :atom\n-   :module\n-   :function\n-   :uuid\n\n_Unknown type always generates ArgumentError exception on compile time._\n\n`module` 'type' means Exop expects a parameter's value to be an atom (a module name) and this module should be already loaded (ready to call it's functions)\n\n`uuid` is not actually a \"type\" but I placed this under `:type` check because there is no reason to have dedicated `:uuid` check.\n\n#### `required`\n\nChecks the presence/absence of a parameter in passed to `run/1` params collection.\nGiven parameters collection fails the validation only if required parameter is missed,\nif required parameter's value is `nil` this parameter will pass this check.\n\n```elixir\nparameter :param_a                   # the same as required: true, required by default\nparameter :param_b, required: false  # this parameter is not required\n```\n\nBy default, a parameter is required (since version 1.2.0, `required: true`).\nIf you want to specify a parameter is not required, provide `required: false`.\nWhy? Because you might find that you repetitively type `required: true` for almost every parameter in a contract. I think if you provide a parameter to an operation (define it in a contract) you expect to get it. Cases, when you need a parameter passed into an operation (and don't really care whether it is present or not), are pretty rare.\n\n_Since version 1.1.0 the behavior of this check has been changed. Check out CHANGELOG for more info._\n\n#### `default`\n\nChecks the presence of a parameter in passed to `run/1` params collection,\nand if the parameter is missed - assign default value to it.\n\n```elixir\nparameter :some_param, default: \"default value\"\n\n# default value can be also a 1-arity function output\nparameter :a, type: :integer, default: &__MODULE__.default_a/1\nparameter :b, type: :integer\n\n# this function takes params given to `run/1`\ndef default_a(params), do: params.b + 1\n\n#iex> YourOperation.run(b: 1)\n#iex> %{a: 2, b: 1}\n```\n\n#### `numericality`\n\nChecks whether a parameter's value is a number and other numeric constraints.\nAll possible constraints are listed in the example below.\n\n```elixir\nparameter :some_param, numericality: %{equal_to: 10, # (aliases: `equals`, `is`, `eq`)\n                                       greater_than: 0, # (alias: `gt`)\n                                       greater_than_or_equal_to: 10 # (aliases: `min`, `gte`),\n                                       less_than: 20, # (alias: `lt`)\n                                       less_than_or_equal_to: 10 # (aliases: `max`, `lte`)}\n```\n\n#### `equals`\n\n(alias: `exactly`)\n\nChecks whether a parameter's value exactly equals given value (with type equality).\n\n```elixir\nparameter :some_param, equals: 100.5\nparameter :some_param, exactly: 100.5\n```\n\n#### `in`\n\nChecks whether a parameter's value is within a given list.\n\n```elixir\nparameter :some_param, in: ~w(a b c)\n```\n\n#### `not_in`\n\nChecks whether a parameter's value is not within a given list.\n\n```elixir\nparameter :some_param, not_in: ~w(a b c)\n```\n\n#### `format`\n\n(alias: `regex`)\n\nChecks wether parameter's value matches given regex.\n\n```elixir\nparameter :some_param, format: ~r/foo/\nparameter :some_param, regex: ~r/foo/\n```\n\n#### `length`\n\nChecks the length of a parameter's value. The value should be one of handled types:\n\n-   list (items count)\n-   string (chars count)\n-   atom (treated as string)\n-   map (key-value pairs count)\n-   tuple (items count)\n\n`length` check is complex as `numericality` (should define map of inner checks).\nAll possible checks are listed in the example below.\n\n```elixir\nparameter :some_param, length: %{gte: 5, gt: 4, min: 5, lte: 10, lt: 9, max: 10, is: 7, in: 5..8}\n```\n\n#### `inner`\n\nChecks the inner of either Map or Keyword parameter. It applies checks described in `inner` map to\nrelated inner items.\n\n```elixir\n# some_param = %{a: 3, b: \"inner_b_attr\"}\n\nparameter :some_param, type: :map, inner: %{\n  a: [type: :integer],\n  b: [type: :string, length: %{min: 1, max: 6}]\n}\n\n# you can omit `type` and `inner` checks keywords in order to check inner of your parameter,\n# when `type` hasn't been specified explicitly, both keyword and map types pass the `type` validation\nparameter :some_param, %{\n  a: [type: :integer],\n  b: [type: :string, length: %{min: 1, max: 6}]\n}\n```\n\nAnd, of course, all checks on a parent parameter (`:some_param` in the example) are still applied.\n\n#### `struct`\n\nChecks whether the given parameter is expected structure.\n\n```elixir\nparameter :some_param, struct: %SomeStruct{}\n# or\nparameter :some_param, struct: SomeStruct\n```\n\n#### `list_item`\n\nChecks whether each of list items conforms defined checks.\nAn item's checks could be any that Exop offers:\n\n```elixir\n# list_param = [\"1234567\", \"7chars\"]\n\n# you can omit `type` check while you're passing a list to an operation\nparameter :list_param, list_item: %{type: :string, length: %{min: 7}}\n```\n\nEven more complex like `inner`:\n\n```elixir\n# list_param = [\n#   %TestStruct{a: 3, b: \"6chars\"},\n#   %TestStruct{a: nil, b: \"7charss\"}\n# ]\n\nparameter :list_param, list_item: %{inner: %{\n                                              a: %{type: :integer},\n                                              b: %{type: :string, length: %{min: 7}}\n                                            }}\n```\n\nMoreover, `coerce_with` and `default` options are available too.\n\n#### `func`\n\nChecks whether an item is valid over custom validation function.\n\nA validation is treated as failed if callback function returns one of results:\n\n-   `{:error, _your_error_message_or_payload}`\n-   `:error`\n-   `false`\n\nEverything else is treaded as successful validation result.\n\nIf the validation function returns either `false` or `:error`, the default message `\"not valid\"` is used in your operation's validation results.\n\nThe validation function receives two arguments:\n\n-   a tuple with a validating parameter's name and value\n-   a map of all parameters given to an operation\n\nThose arguments allow you to make a parameter validations which depend on other parameters and their values.\n\n```elixir\nparameter :some_param, func: &__MODULE__.your_validation/2\n\n@spec your_validation({atom() | String.t(), any()}, map()) :: any()\ndef your_validation({param_name, param_value}, all_received_params_map) do\n  # your validation logic based on given arguments is here\nend\n```\n\n_it is possible to combine :func check with others (though not preferable), just make sure this check is the last check in the list_\n\n##","ref":"readme.html#parameter-checks","title":"Parameter checks - README","type":"extras"},{"doc":"It is a parameter attribute which allow you to have other checks for a parameter whilst have a possibility to pass `nil` as the parameter's value.\nIf `allow_nil: true` and `nil` is passed as a parameter value _all_ the parameter's checks are ignored during validation.\n\n```elixir\ndefmodule YourOperation do\n  use Exop.Operation\n\n  parameter :a, type: :integer, allow_nil: true\n  parameter :b, type: :integer, allow_nil: false\n\n  def process(params), do: params\nend\n\n{:ok, %{a: 1, b: 1}} = YourOperation.run(a: 1, b: 1)\n{:ok, %{a: nil, b: 1}} = YourOperation.run(a: nil, b: 1)\n{:error, {:validation, %{b: [\"doesn't allow nil\"]}}} = YourOperation.run(a: nil, b: nil)\n```\n\n_By default (if you omit `allow_nil`), a parameter is treated as `allow_nil: false`_\n\n##","ref":"readme.html#allow_nil","title":"allow_nil - README","type":"extras"},{"doc":"This option allows you to pass a parameter to `run/1` and `run!/1` functions with one name and work with this parameter within an operation under another name.\n\n```elixir\ndefmodule YourOperation do\n  use Exop.Operation\n\n  parameter :a, type: :integer, from: \"a\"\n  parameter :b, type: :string, from: :bB\n\n  def process(params), do: params\nend\n\n# now you can invoke YourOperation with such params:\n#   Youroperation.run(%{\"a\" => 1, b: \"1\"})\n#   Youroperation.run(%{a: 1, bB: \"1\"})\n#   Youroperation.run(%{\"a\" => 1, bB: \"1\"})\n# and get:\n#   {:ok, %{a: 1, b: \"1\"}}\n```\n\nThe same works for parameters given as a Keyword as well (in this case `:from` value should be an atom):\n\n```elixir\ndefmodule YourOperation do\n  use Exop.Operation\n\n  parameter :a, type: :integer, from: :aA\n  parameter :b, type: :string, from: :bB\n\n  def process(params), do: params\nend\n\n# Youroperation.run(a: 1, b: \"1\")\n# Youroperation.run(aA: 1, bB: \"1\")\n# {:ok, %{a: 1, b: \"1\"}}\n```\n\nWhy? Simply because sometimes you're not in control of incoming parameters but don't want to map them each time you need to use'em by yourself (good example: params in Phoenix controller's action, which come as a map with string keys).\n\n_This option doesn't work for `:inner` check's inner parameters currently._\n\n##","ref":"readme.html#from","title":"from - README","type":"extras"},{"doc":"Checks whether a parameter's value (list) is a subset of a defined check-list.\nTo pass this check, all items within given into an operation parameter should be included into check-list,\notherwise the check is failed.\n\n```elixir\nparameter :some_param, subset_of: [1, 2, :a, \"b\", C]\n\n# {:ok, _} = MyOperation.run(some_param: [1, :a, C])\n# {:ok, _} = MyOperation.run(some_param: [:a])\n# {:error, _} = MyOperation.run(some_param: [])\n# {:error, _} = MyOperation.run(some_param: [3, :a, C])\n```\n\n#","ref":"readme.html#subset_of","title":"subset_of - README","type":"extras"},{"doc":"In some cases you might want to make an 'early return' from `process/1` function.\nFor this purpose you can call `interrupt/1` function within `process/1` and pass an interruption reason to it.\nAn operation will be interrupted and return `{:interrupt, your_reason}`\n\n```elixir\n# ...\ndef process(_params) do\n  interrupt(%{fail: \"oops\"})\n  :ok # will not return it\nend\n# ...\n\nSomeOperation.run(a: 1) # {:interrupt, %{fail: \"oops\"}}\n```\n\n_`run!/1` invocation doesn't affect interruption result: {:interrupt, \\_your_result} tuple will be returned anyway as expected and handled result._\n\n#","ref":"readme.html#interrupt","title":"Interrupt - README","type":"extras"},{"doc":"It is possible to coerce a parameter before the contract validation, all validation checks\nwill be invoked on coerced parameter value.\nSince coercion changes a parameter before any validation has been invoked,\ndefault values are resolved (with `:default` option) before the coercion.\nThe flow looks like: `Resolve param default value -> Coerce -> Validate coerced`\n\nIf coercion function returns an error tuple it will be treated as validation failure: an operation's invokation stops and that error tuple will be returned as a result.\n\n`:coerce_with` accepts 2-arity function. This function takes a tuple `{coerced_param_name, coerced_param_value}` as the first argument and a map with all the parameters that have been passed to either `run/1` or `run!/1` function as the second argument.\n\n```elixir\nparameter :a, type: :integer\nparameter :b, type: :string, coerce_with: &__MODULE__.to_string/2\nparameter :c, type: :string, coerce_with: &__MODULE__.to_string/2\n\ndef to_string({_, value}, %{a: _, b: _, c: _} = _received_params) when is_integer(value) do\n  Integer.to_string(c_value)\nend\n\ndef to_string({_, value}, _received_params) when is_binary(c_value) do\n  value\nend\n\ndef to_string({:c, c_value}, _received_params) do\n  # special coercion for :c parameter here\nend\n```\n\n_Why it is so? Because there are cases when you can use the same coercion function for multiple params and/or coerce a parameter depending on another's value and need to have all this information._\n\n#","ref":"readme.html#coercion","title":"Coercion - README","type":"extras"},{"doc":"It is possible to define a policy that will be used for authorizing the possibility to invoke an operation. So far, there is a simple policy implementation and usage.\n\nJust define a module with a bunch of functions, each takes a single argument (any type) and returns true/false (authorization result) or any term (which will be represented as authorization error reason.\n\n```elixir\n  defmodule MonthlyReportPolicy do\n    def can_read?(%{user_role: \"admin\"}), do: true\n    def can_read?(\"admin\"), do: true\n    def can_read?(%User{role: \"manager\"}), do: true\n    def can_read?(:manager), do: true\n    def can_read?(_opts), do: false\n\n    def can_write?(%{user_role: \"admin\"}), do: true\n    def can_write?(%{user_role: \"manager\"}), do: false\n    def can_write?(_), do: [:your, \"reason\"] # (the result error will be: {:error, {:auth, [:your, \"reason\"]}})\n  end\n```\n\nIn this policy two actions (checks) defined (`can_read?/1` & `can_write?/1`).\nEvery action expects an argument for a check. It's up to you how to handle this argument and turn it into the actual check.\n\n-   next step - link an operation and a policy\n\n```elixir\n  defmodule ReadOperation do\n    use Exop.Operation\n\n    policy MonthlyReportPolicy, :can_read?\n\n    parameter :user, struct: User\n\n    def process(%{user: %User{} = user}) do\n      # make some reading...\n    end\n  end\n```\n\n-   finally - call `authorize/1` within `process/1`\n\n```elixir\n  defmodule ReadOperation do\n    use Exop.Operation\n\n    policy MonthlyReportPolicy, :can_read?\n\n    parameter :user, struct: User\n\n    def process(params) do\n      authorize(params.user)\n\n      # make some reading...\n    end\n  end\n```\n\n_Please, note: if authorization fails, any code after (below) auth check\nwill be postponed (an error `{:error, {:auth, _policy_action_name}}` will be returned immediately as an operation result)_\n\n#","ref":"readme.html#policy-check","title":"Policy check - README","type":"extras"},{"doc":"If you'd like to handle various operations fails with a certain logic (for example log it into Graylog)\nyou can use `Exop.Fallback`.\n\nDefine a fallback module:\n\n```elixir\n  defmodule FallbackModule do\n    use Exop.Fallback\n\n    def process(operation, params, error) do\n      # your error handling code here\n      :some_fallback_result\n    end\n  end\n```\n\nhere you need to define and implement `process/3` function which takes following params:\n\n-   failed operation module\n-   params that were passed into the operation\n-   an error result which was returned by the operation\n\nUse your fallback in operations like this:\n\n```elixir\ndefmodule SomeOperation do\n  use Exop.Operation\n\n  fallback FallbackModule, return: true\n\n  parameter :a, type: :integer\n  parameter :b, type: :integer\n\n  def process(%{a: a, b: b}), do: a + b\nend\n```\n\nThe results of the operation executions:\n\n```elixir\n# SomeOperation will fail\niex> SomeOperation.run(a: 1, b: \"2\")\n:some_fallback_result\n\n# SomeOperation will be successful\niex> SomeOperation.run(a: 1, b: 2)\n{:ok, 3}\n```\n\nDuring a fallback definition you can add `return: true` option so in the example case\n`SomeOperation.run/1` will return the result of the fallback (`FallbackModule.process/3`\nfunction's result - `:some_fallback_result`).\nIf you want `SomeOperation.run/1` to return original result (`{:error, {:validation, %{a: [\"has wrong type\"]}}}`)\nspecify `return: false` option or just omit it in a fallback definition.\n\n#","ref":"readme.html#fallback-module","title":"Fallback module - README","type":"extras"},{"doc":"If you'd like to handle a side effect successful operation (for example brodcast results to Phoenix.PubSub)\nyou can use `Exop.Callback`.\n\nDefine a callback module:\n\n```elixir\n  defmodule CallbackModule do\n    use Exop.Callback\n\n    def process(operation, params, result, opts) do\n      Phoenix.PubSub.broadcast(MyApp.PubSub, :topic, result)\n    end\n  end\n```\n\nhere you need to define and implement `process/4` function which takes following params:\n\n-   operation module\n-   params that were passed into the operation\n-   an successful result which was returned by the operation\n-   opts keyword list for additional metadata\n\nUse your callback in operations like this:\n\n```elixir\ndefmodule SomeOperation do\n  use Exop.Operation\n\n  callback CallbackModule\n\n  parameter :a, type: :integer\n  parameter :b, type: :integer\n\n  def process(%{a: a, b: b}), do: a + b\nend\n```\n\nYou can use `opts` keyword list to add needed metadata to your call like this:\n\n```elixir\n  defmodule CallbackModule do\n    use Exop.Callback\n\n    def process(operation, params, result, opts) do\n      Phoenix.PubSub.broadcast(MyApp.PubSub, opts[:topic], result)\n    end\n  end\n```\n\nand then use the callback like this\n\n```elixir\ndefmodule SomeOperation do\n  use Exop.Operation\n\n  callback CallbackModule, topic: :some_topic\n\n  parameter :a, type: :integer\n  parameter :b, type: :integer\n\n  def process(%{a: a, b: b}), do: a + b\nend\n```","ref":"readme.html#callback-module","title":"Callback module - README","type":"extras"},{"doc":"As said earlier, operations in most cases called by `run/1` function. This function\nreceives parameters collection. It's not required to pass to `run/1` function parameters\nonly described in the operation's contract, but only described parameters will be validated and gived to `process/1` function, all other will be filtered out from the process.\n\n_filtering out parameters which are not defined in a contract is here to support the whole idea of defining the right operation's contract and take care of what you really need for a certain business process / function_\n\n`run/1` function validates received parameters over the contract and if all parameters passed\nthe validation, the `run/1` function calls the code defined in `process/1` function.\n\n```elixir\niex> SomeOperation.run(param1: 1, param2: \"2\")\n_some_result_\n```\n\nIf at least one of the given parameters didn't pass the validation `process/1` function's code\nwill not be invoked and corresponding warning in the application's log will appear.\n\nThere is \"bang\" version of `run/1` exists. Function `run!/1` does the same things that its sibling does,\nthe only difference is a result of invocation, it might be:\n\n-   if a contract validation passed - the actual result of an operation (result of a code, described in `process/1`)\n-   if a contract validation failed - an error `Exop.Validation.ValidationError` is raised\n-   if an operation returns an error tuple - an error `Exop.Operation.ErrorResult` is raised\n-   in case of manual interruption - `{:interrupt, _reason}`\n\n_You always can bypass the validation simply by calling `process/1` function itself, if needed._","ref":"readme.html#operation-invocation","title":"Operation invocation - README","type":"extras"},{"doc":"If received parameters passed a contract validation, a code defined in `process/1` will be invoked.\nOr you will receive `@type validation_error :: {:error, :validation_failed, map()}` as a result otherwise.\n`map()` as errors reasons might look like this:\n\n```elixir\n%{param1: [\"has wrong type\"], param2: [\"is required\", \"must be equal to 3\"]}\n```\n\nAn operation can return one of results listed below (depends on passed in params and operation definition):\n\n-   an operation was completed successfully:\n    -   `{:error, _your_error_reason_}` (if an :error-tuple (any length, but `:error` atom should be the first element) was returned by `process/1` function)\n    -   `{:ok, any()}` (otherwise, even if `{:ok, _your_result_}` tuple was returned by `process/1` function)\n-   a contract validation failed: `{:error, {:validation, map()}}`\n-   if `interrupt/1` was invoked: `{:interrupt, any()}`\n-   policy check failed:\n    -   `{:error, {:auth, :undefined_policy}}`\n    -   `{:error, {:auth, :undefined_action}}`\n    -   `{:error, {:auth, atom()}}`\n\n_For the \"bang\" version of `run/1` see results description above._","ref":"readme.html#operation-results","title":"Operation results - README","type":"extras"},{"doc":"Sometimes you need to aggregate/group 'atom' operations into a single one operation responsible for\nsome complex business process/logic.\nYou have a few approaches to do it (`with` for example) but mb you'll find `Exop.Chain` more handy.\n\n`Exop.Chain` provides a simple way to organize a number of Exop.Operation modules into an invocation chain.\n\n```elixir\ndefmodule CreateUser do\n  use Exop.Chain\n\n  alias Operations.{User, Backoffice, Notifications}\n\n  operation User.Create\n  operation Backoffice.SaveStats\n  operation Notifications.SendEmail\n\n  # or you can use step alias:\n  #   step User.Create\n  #   step Backoffice.SaveStats\n  #   step Notifications.SendEmail\nend\n```\n\nThis is how invoke this chain:\n\n```elixir\niex> CreateUser.run(name: \"User Name\", age: 37, gender: \"m\")\n```\n\n`Exop.Chain` defines `run/1` function under the hood (like common operations do) that accepts `keyword()`, `map()` or `struct()` as params.\nThose params will be passed into the first operation in the chain.\nBear in mind that each of chained operations (except the first one) awaits a returned result of\na previous operation as incoming params.\n\nSo in the example above `CreateUser.run(name: \"User Name\", age: 37, gender: \"m\")` will invoke\nthe chain by passing `[name: \"User Name\", age: 37, gender: \"m\"]` params to the first `User.Create` operation.\nThe result of `User.Create` operation will be passed to `Backoffice.SaveStats`\noperation as its params and so on.\n\nOnce any of operations in the chain returns non-ok-tuple result (error result, interruption, auth error etc.)\nthe chain execution interrupts and error result returned (as the chain (`CreateUser`) result).\n\n#","ref":"readme.html#operations-chain","title":"Operations chain - README","type":"extras"},{"doc":"You can pass additional parameters to any operation in a chain (with either an exact value or 0-arity function):\n\n```elixir\ndefmodule CreateUser do\n  use Exop.Chain\n\n  alias Operations.{User, Backoffice, Notifications}\n\n  operation User.Create\n  operation Backoffice.SaveStats, logger: MyFancyLoggerModule\n  # or\n  operation Backoffice.SaveStats, logger: &__MODULE__.logger/0\n  operation Notifications.SendEmail\n\n  def logger, do: MyFancyLoggerModule\nend\n```\n\n#","ref":"readme.html#additional-parameters","title":"Additional parameters - README","type":"extras"},{"doc":"`use Exop.Chain` can take `:name_in_error` option, when it is set to `true` a failed operation in a chain returns the operation's module name as the first elements of output tuple `{YourOperation, {:error, _}}`\n\n```elixir\ndefmodule YourChain do\n  use Exop.Chain, name_in_error: true\n\n  operation Operation1\n  operation Operation2Fail\n  operation Operation3\nend\n\niex> YourChain.run(a: \"1\", b: 2)\n{Operation2Fail, {:error, {:validation, %{a: [\"has wrong type\"]}}}}\n```\n\n`name_in_error: true` doesn't affect operations with a fallback defined (unmodified fallback result is returned).\n\n#","ref":"readme.html#descriptive-errors","title":"Descriptive errors - README","type":"extras"},{"doc":"It is possible to define an invokation condition for an operation in a chain. Meaning if the condition is truthy - the operation will be invoked.\n\nThe condition can be defined with `if: your_func/1` option given to an operation.\n\n```elixir\ndefmodule YourChain do\n  use Exop.Chain\n\n  # operation/step - they are synonims in the context of Exop.Chain\n  operation Operation1\n  operation MultiplyByHundred, if: &__MODULE__.is_it_good_to_go?/1\n  operation DivisionByTen\n\n  def is_it_good_to_go?(previous_operation_output) do\n    # here your condition logic which should return a boolean\n  end\nend\n```\n\nA condition function receives a single argument - the previous operation's output (the second element of `{:ok, _}` tuple, not the tuple itself), turned into map (if the output is Keyword list) so it is easier to pattern-match.\n\nAn operation is invoked if a condition function returns `true`, otherwise the operation won't be invoked.\n\nAnd of course a chain invokation interrupts if the previous operation's result wasn't successful (is not `{:ok, _}` tuple). The previous operation's result is not changed in this case.\n\n#","ref":"readme.html#conditional-operations","title":"Conditional operations - README","type":"extras"},{"doc":"If for some reason you need to change incoming parameters (which are the previous operation result) for your operation(-s) in the chain you can do it with `:coerce_with` option. This option is provided for a particular operation (step) and refers to a 1-arity callback function. This single argument is a map of incoming parameters (the previous operation result). It is converted into a map even if the previous operation returns a keyword list.\n\n```elixir\ndefmodule YourChain do\n  use Exop.Chain\n\n  operation Sum\n  operation MultiplyByHundred, coerce_with: &__MODULE__.coerce/1\n  operation DivisionByTen\n\n  def coerce(%{a: a} = params), do: %{params | a: a * 10}\nend\n```\n\nThe coercion works right before any validation/invokation (how it is with the regular operation). It means that first of all the previous operation's result is coerced and then everything else happens based on those coerced parameters.\n\nCoercion works well with other options like additional parameters and conditional invocation.\n\nHere is the order of an operation handling in a chain:\n\n1. incoming parameters coercion\n2. additional parameters are added\n3. invocation condition is checked","ref":"readme.html#incoming-parameters-coercion","title":"Incoming parameters coercion - README","type":"extras"},{"doc":"Copyright © 2016 - 2020 Andrey Chernykh ( andrei.chernykh@gmail.com )\n\n    This work is free. You can redistribute it and/or modify it under the\n    terms of the MIT License. See the LICENSE file for more details.","ref":"readme.html#license","title":"LICENSE - README","type":"extras"}]}